'use strict';

/*
    This is the core class for this app.
*/

/* global IMatchModals, OLMapModule,GMapModule,MapFind,MapLocations,IMatchMapModals,MapModals,CameraDatabase,MapUtility,IMatchTrackLog,TrackLogConfig,IMatchTracks,KMLExport,moment */

import CameraDatabase from '/system/imatch/camera-db/camera-db.js';

window.appCore = new function() {

    let self = this;

    let imatchInfo = null;

    let mapModule = null;

    let activeOverlay = '';

    let findResultWindowInstanceId = 0;

    let googleSharedKeyWarningShown = false;

    this.camDB = new CameraDatabase()
   
    self.userName = '';
    self.dbReadOnly = false;
    self.specialFolders = {};

    // data storage keys. We replace {u} with the name of the
    // current IMatch user to make the configuration per-user.
    let DATA_KEY_CONFIG = 'imatch.mapapp.{u}.config';

    // Current config version
    const APPCONFIG_VERSION = 6;

    this.appConfig = {
        version : APPCONFIG_VERSION,

        // The selected map type
        mapType : 'osm',
        
        // Options
        showFileData : false,
        showImages: false,
        showLocations: true,
        followFocus: true,

        showFileTrack: false,

        // Save for all map types
        zoomLevel : 12,

        // The initial map center (London), updated by all
        // map modules when they save their configuration
        mapCenter: {
            lng: -0.136357,
            lat: 51.510121
        },

        // The initial find text, as a hint for the user.
        findText: 'london',
        findTarget: '-0.136357, 51.510121',

        // If this is true, markers on the map cannot be moved (except the target marker)
        lockMove: false,

        // The tool to use for selects: circle, rect, poly
        selectTool: 'circle',

        // Use shown coordinates for the select Tool?
        selectToolUseShownCoordinates: false,

        // Where to apply the select filter: scope or db
        selectScope: 'scope',

        trackLog: {
            // UTC offset for all track timestamps.
            // In hours (-0.5 = -00:30).
            trackTimeShift: 0.0,

            // Ignore files with existing GPS data when mapping
            ignoreFilesWithGPS: true,

            // The 5 last loaded GPS files
            files: [],

            folders: [],

            rememberFolder: false
        },

        // Version 2+
        // Show field of view? Possible values: {off,focus,all}
        fieldOfView: 'off',

        // Version 3+

        // Did the user acknowledge the provacy statement?
        privacyPolcyAgreement: false
    }

    // Per-session options
    this.sessionSettings = {
        showApplyTargetMarkerPromt : true
    }


    // These options are retrieved from IMatch.
    // They represent the settings in Edit > Preferences > Geo & Maps
    this.IMatchConfig = {

        // API Keys
        geoNamesKey: '',
        gmapsKey: '',
        bingMapsKey: '',
        HEREMapsKey: '',
        
        // The locale to use for map lookups. 
        locale: 'en',

        // The provider to use for searches (1=geonames, 10=google)
        searchProvider: 1,
        
        // The provider to use for reverse geo-coding (1=geonames, 10=google)
        codingProvider: 1
    }

    let mapTypes = []

    let currentToolbar = 'default';

    let tracks = new IMatchTracks;

    // This is used while we geo-code files and update data in IMWS.
    let geoCodeTaskData = {
        running : false,
        tasks : [],
        index : 0,
        filesUpdated : [],
        overlay : null
    }

    let cameraDBWarningShown = false;



    /////////////////////////////////////////////////////////////////////////////
    // This kicks off  everything
    $(document).ready(function () {
        window.appCore.init();
    });


    /////////////////////////////////////////////////////////////////////////////
    // This function kicks of everything
    this.init = function() {

        $('#btn-locate-target').attr('disabled',true);

        // Translate the app, then initialize everything
        IMatchTranslate.init(IMatchTranslate.getUserLanguage(),'en','res/').then(function() {
        // IMatchTranslate.init('de','en','res/').then(function() {

            IMatch.userInfo().then(response => {
                IMatch.localeInitMomentJS(response);
            });

            // Get more detailed info about IMatch
            IMatch.get(`v1/imatch/info`).then(function(response) {
                imatchInfo = response;

                // We load the configuration or setup defaults first
                loadConfig().then(function() {
                    
                    // Setup all available maps
                    setupMapConfig();

                    initMenu();

                    // Show the Privacy Policy modal before continuing.                    
                    if (!self.appConfig.privacyPolcyAgreement) {
                        var m = new IMatchModals();
                        m.showMessageBoxYesNo({
                            headline: IMatchTranslate.getStr('MSGBOX_PRIVPOL_TITLE'),
                            content: IMatchTranslate.getStr('MSGBOX_PRIVPOL_TEXT'),
                            okCaption: IMatchTranslate.getStr('MSGBOX_PRIVPOL_OK'),
                            cancelCaption: IMatchTranslate.getStr('MSGBOX_PRIVPOL_CANCEL'),
                            icon: 'fa fa-user-circle fa-3x',
                            okCallback: function() {
                                self.appConfig.privacyPolcyAgreement = true;
                                self.saveConfig();
                                self.initMap();
                            },
                            cancelCallback: function()  {
                                $('#app-container').hide();
                            }
                        });
                    }
                    else {
                        self.initMap();
                    }

                    // 2020.11+

                    if (!self.appConfig.showFileTrack) {
                        self.appConfig.showFileTrack = false;
                    }
                });

                openWebSockets();
            });
            
        });

        // Keyboard shortcuts

        $(document).keydown(function(event) {

            // console.log(event.which);
            if(event.which === 70) { // F
                if (activeOverlay == '') {
                    event.preventDefault();
                    $('#btn-find').click();
                }
            }
            else if(event.which === 76) { // L
                if (activeOverlay == '') {
                    $('#btn-location-goto').click();
                }
            }
            else if (event.which === 75) { // K
                if (activeOverlay == '') {
                    $('#btn-kml-export').click();
                }
            }
            else if(event.which === 84) { // T
                if (activeOverlay == '') {
                    $('#btn-track-load').click();
                }
            }
            else if(event.which === 79) { // O
                if (activeOverlay == '') {
                    $('#btn-track-load-folder').click();
                }
            }
            
            else if(event.which === 82) { // R
                if (activeOverlay == '') {
                    $('#check-show-file-track').click();
                }
            }
            else if(event.which === 90) { // Z
                if (event.ctrlKey) {
                    self.undo();
                }
            }
           
        });
    }

    /**
     * Returns true if the app is running in debug mode.
     * @returns {Boolean}
     */
    self.debugMode = function()
    {
        return imatchInfo != null ? imatchInfo.developerMode : false;
    }

    /////////////////////////////////////////////////////////////////////////////
    //    
    self.initMap = function()
    {
        var mt = self.getMapConfig(self.appConfig.mapType);
        if (mt != null) {
            loadMapTemplate(mt);
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    //    
    self.undo = function()
    {
        closeModals();
        if (mapModule) {
            mapModule.onUndo();
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Open web sockets and process messages sent by IMatch and IMWS.
    function openWebSockets()
    {
        // Open the web socket for IMatch
        if (IMatch.isEmbedded()) {
            IMatch.openWebSocket();
        }
        else {
            IMatch.openWebSocket(IMatch.DEFAULT_WS_URL);
        }

        // De-bounce helper
        var delayLoadFiles = IMatch.debounce(function() {
            if (mapModule) {
                mapModule.loadFiles();
            }
        },500);
        
        IMatch.webSocket().onmessage = function(e) {
            if (mapModule) {
                var msg = JSON.parse(e.data);
                //console.log(msg);
                // Reload the files when the selection in the file window changes.
                if ((msg.msg == 'fileWindow.selectionChanged') ||
                    (msg.msg == 'fileWindow.scopeChanged') ||
                    (msg.msg == 'fileWindow.selectionCleared')
                ) {
                    delayLoadFiles();
                }
                else if (msg.msg == 'app.globalOptionsChanged') {
                    self.loadIMatchSettings();
                }
            }
        }

        // Open the web socket for IMWS
        if (IMatch.isEmbedded()) {
            IMWS.openWebSocket();
        }
        else {
            IMWS.openWebSocket(IMatch.DEFAULT_WS_URL);
        }
      
        // Process
      
        IMWS.webSocket().onmessage = function(e) {
            var msg = JSON.parse(e.data);
            switch (msg.ma) {

            // A database object was updated.
            case IMWS.EVENT_CODES.evObjectUpdate:
                // Depending on the type of update, we may need
                // to update the map.
                switch (msg.mi) {
                case IMWS.EVENT_CODES.evOUFileUpdate:
                case IMWS.EVENT_CODES.evOUFileRename:
                case IMWS.EVENT_CODES.evOUThumbUpdate:
                case IMWS.EVENT_CODES.evOUThumbSetUpdate:
                case IMWS.EVENT_CODES.evOUMBFileValueUpdate:
                case IMWS.EVENT_CODES.evOUMBLabelUpdate:
                case IMWS.EVENT_CODES.evOUMBRatingUpdate:
                    // All these events may impact our map, if the 
                    // updated file is currently shown on the map.
                    if (mapModule) {
                        mapModule.onFileUpdate(msg);
                    }
                    break;
                case IMWS.EVENT_CODES.evOUGeoLocationsUpdate:
                    // Geo locations were updated (added, removed or modified).
                    if (mapModule) {
                        mapModule.onLocationUpdate(msg);
                    }
                    break;
                }
                break;
            }
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Setup drag and drop handling for GPX files
    /*
    function initDragAndDrop()
    {
        // We use the track button
        var dropZone = $('#app-container');

        dropZone.on('dragenter',function(e) {
            // Allow drop
            e.preventDefault(); 
            dropZone.addClass('drop-hot');
        });

        dropZone.on('dragover',function(e) {
            e.preventDefault(); // Allow drop
        });

        dropZone.on('dragleave',function(e) {
            dropZone.removeClass('drop-hot');
            e.preventDefault(); // Allow drop
        });


        // Dropped (released the mouse)
        dropZone.on('drop',function(e) {
            dropZone.removeClass('drop-hot');
            e.preventDefault();

            for (var i = 0; i < e.originalEvent.dataTransfer.types.length; ++i) {
                var typeName = e.originalEvent.dataTransfer.types[i];

                // If the typeName starts with text we check for an imatchDrop
                // object in JSON format. This is used when you drag something from
                // IMatch into the web browser. It is the 'richest' format.
                
                if (typeName.match(/^TEXT/i)) {  // case-insensitive compare, must start with TEXT
                    var text = e.originalEvent.dataTransfer.getData(typeName);

                    // We do a check to see if this is a drop object from IMatch.
                    // This is the 'richest' format with the most detailed data.
                    if (text.match(/imatchDrop/i)) {
                        // Yes. Parse it into a real JavaScript object.
                        // And then pretty-print it
                        var obj = JSON.parse(text);
                    }
                    else {
                        // Text format, but not with IMatch data
                        var text = e.originalEvent.dataTransfer.getData(typeName);
                    }
                }
                // Do we have files
                else if (e.originalEvent.dataTransfer.files.length > 0) {
                    for (var f = 0; f < e.originalEvent.dataTransfer.files.length; ++f) {
                        console.log(e.originalEvent.dataTransfer.files[f].name);
                    }
                }
            }
        });        
    }
    */

    /////////////////////////////////////////////////////////////////////////////
    // Clean up stuff when the window closes or the browser navigates
    $(window).on("unload", function() {
        if (mapModule) {
            mapModule.onDestroy();
        }
        self.saveConfig();
    });

    // The ids of the IMatch settings we use in this app (geo).
    this.IMSETTINGS = {
        GoogleMapsAPIKey : 110,
        GeoNamesAPIKey : 210,
        epBingMapsAPIKey : 510,
        epHEREMapsAPIKey : 560,
        epGeoSearchService : 2000,
        epGeoCodingService: 2010,
        epGeoLanguage : 2020
    }

    /////////////////////////////////////////////////////////////////////////////
    // Load the configuration from the data store.
    // This returns a promise because it needs to make some IMWS/IMatch endpoint calls.
    function loadConfig() {
        var dfd = jQuery.Deferred();

        // Request the info first. We need the user name to load all other settings.
        IMWS.get('info').then(response => {

            self.userName = response.userName;

            self.dbReadOnly = response.database.readOnly;

            // The per-user data storage key
            DATA_KEY_CONFIG = DATA_KEY_CONFIG.replace('{u}',self.userName);

            self.camDB.init();

            // Retrieve the settings from Geo & Maps and copy them into IMatchConfig

            var ids = [];
            $.each(self.IMSETTINGS,function(k,v) {
                ids.push(v);
            }, this);

            // We use when to wait for multiple promises. Easym thanks to jQuery ;-)

            $.when(
                self.loadIMatchSettings(),
            
                // Load the configuration data from the last session
                IMWS.dataGet({name: DATA_KEY_CONFIG}).then(function(response) {
                    var data = JSON.parse(response.value);

                    // Check the version
                    if (data.version == self.appConfig.version) {
                        self.appConfig = data;
                    }
                    else {
                        // Upgrade older config versions
                        if (data.version == 1) {
                            // 1 => 2
                            self.appConfig = data;
                            self.appConfig.version = APPCONFIG_VERSION;

                            // Version 2+
                            self.appConfig.fieldOfView = 'focus';
                            self.appConfig.privacyPolcyAgreement = false;
                        }
                        else if (data.version == 2) {
                            // 2 => 3
                            self.appConfig = data;
                            self.appConfig.version = APPCONFIG_VERSION;

                            self.appConfig.privacyPolcyAgreement = false;
                        }
                        else if (data.version == 3) {
                            // 3 => 4
                            self.appConfig = data;
                            self.appConfig.version = APPCONFIG_VERSION;

                            self.appConfig.selectToolUseShownCoordinates = false;
                        }
                        else if (data.version < 5) {
                            // 4 => 5 
                            self.appConfig = data;
                            self.appConfig.version = APPCONFIG_VERSION;

                            self.appConfig.trackLog.folders = [];
                            self.appConfig.trackLog.rememberFolder = false;
                        }
                    }
                }),

                // We need to know the location of some of the special folders...

                IMWS.get('v1/filesystem/specialfolders').then(function(response) {
                    self.specialFolders = response.folders;
                })
            ).always(function() {
                // This is called when all outstanding promises above are resolved and
                // we have loaded everything.
                IMatch.addTelemetryEntry({
                    type: 'metric',
                    name: 'app.mapapp.provider',
                    text: self.appConfig.mapType
                });
                dfd.resolve();
            });
        });

        return dfd.promise();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Load some global IMatch settings, like
    this.loadIMatchSettings = function()
    {
        var dfd = $.Deferred();

        var ids = [];
        $.each(this.IMSETTINGS,function(k,v) {
            ids.push(v);
        });

        // Access some internal IMatch properties (from Geo & Maps)
        IMatch.getIMatchSetting('geo',ids.join(',')).then(settings => {
            this.IMatchConfig.geoNamesKey = settings[this.IMSETTINGS.GeoNamesAPIKey].value;
            this.IMatchConfig.gmapsKey = settings[this.IMSETTINGS.GoogleMapsAPIKey].value;
            this.IMatchConfig.bingMapsKey = settings[this.IMSETTINGS.epBingMapsAPIKey].value;
            this.IMatchConfig.HEREMapsKey = settings[this.IMSETTINGS.epHEREMapsAPIKey].value;
            this.IMatchConfig.locale = settings[this.IMSETTINGS.epGeoLanguage].value;
            // For these we use the numeric value
            this.IMatchConfig.searchProvider = settings[this.IMSETTINGS.epGeoSearchService].data;
            this.IMatchConfig.codingProvider = settings[this.IMSETTINGS.epGeoCodingService].data;
            dfd.resolve();
        },function(error) {
            dfd.reject(error);
        });

        return dfd.promise();
    }

    /////////////////////////////////////////////////////////////////////////////
    // Save the configuration data.
    this.saveConfig = function()
    {
        self.appConfig.version = APPCONFIG_VERSION;

        IMWS.dataSet({
            name: DATA_KEY_CONFIG,
            value: JSON.stringify(self.appConfig)
        });
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    this.tracks = function()
    {
        return tracks;
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    this.getMapModule = function()
    {
        return mapModule;
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Setup the available map types. 
    // Bing and OSM are handled by OpenLayers. 
    // Google Maps uses a separate implementation.
    function setupMapConfig()
    {
        mapTypes = [
            {
                type: 'osm',
                module: 'ol',
                group: 'OpenStreetMap',
                // The template to load at runtime
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_OSM'),
                maxZoom: 19,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'stamen',
                module: 'ol',
                group: 'Stamen',
                // The template to load at runtime
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_STAMEN'),
                maxZoom: 19,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            // Bing

            {
                type: 'bing-aerial',
                module: 'ol',
                group: 'Bing',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_BM_AERIAL'),
                maxZoom: 19,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            {
                type: 'bing-aeriallbl',
                module: 'ol',
                group: 'Bing',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_BM_AERIALLBL'),
                maxZoom: 19,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            {
                type: 'bing-road',
                module: 'ol',
                group: 'Bing',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_BM_ROAD'),
                maxZoom: 19,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            {
                type: 'bing-ordnanceSurvey',
                module: 'ol',
                group: 'Bing',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_BM_ORDNANCE'),
                maxZoom: 19,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            
            // DOES NOT WORK (yet?)
            // {
            //     type: 'bing-birdseye',
            //     module: 'ol',
            //     group: 'Bing',
            //     template: 'ol/ol-tpl.html',
            //     name: IMatchTranslate.getStr('PROVIDER_BM_BIRDSEYE'),
            //     maxZoom: 19,
            //     params: {
            //         fields: 'id,gps,name',
            //         imagesize: 'T'
            //     }
            // },

            // HERE

            {
                type: 'here-normal-day',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_STREET_DAY'),
                maxZoom: 20,
                layerData: {
                    base: 'base',
                    type: 'maptile',
                    scheme: 'normal.day',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'here-normal-night',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_STREET_NIGHT'),
                maxZoom: 20,
                layerData: {
                    base: 'base',
                    type: 'maptile',
                    scheme: 'normal.night',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'here-pedestrian-day',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_PEDESTRIAN_DAY'),
                maxZoom: 20,
                layerData: {
                    base: 'base',
                    type: 'maptile',
                    scheme: 'pedestrian.day',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            {
                type: 'here-pedestrian-night',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_PEDESTRIAN_NIGHT'),
                maxZoom: 20,
                layerData: {
                    base: 'base',
                    type: 'maptile',
                    scheme: 'pedestrian.night',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'here-terrain-day',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_TERRAIN_DAY'),
                maxZoom: 20,
                layerData: {
                    base: 'aerial',
                    type: 'maptile',
                    scheme: 'terrain.day',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            {
                type: 'here-satellite-day',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_SATELLITE_DAY'),
                maxZoom: 20,
                layerData: {
                    base: 'aerial',
                    type: 'maptile',
                    scheme: 'satellite.day',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },
            {
                type: 'here-hybrid-day',
                module: 'ol',
                group: 'HERE',
                template: 'ol/ol-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_HERE_HYBRID_DAY'),
                maxZoom: 20,
                layerData: {
                    base: 'aerial',
                    type: 'maptile',
                    scheme: 'hybrid.day',
                },
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            // Google Maps

            {
                type: 'google-satellite',
                module: 'gm',
                group: 'Google',
                template: 'gm/gmaps-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_GM_SAT'),
                maxZoom: 21,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'google-street',
                module: 'gm',
                group: 'Google',
                template: 'gm/gmaps-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_GM_STREET'),
                maxZoom: 21,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'google-hybrid',
                module: 'gm',
                group: 'Google',
                template: 'gm/gmaps-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_GM_HYBRID'),
                maxZoom: 21,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            },

            {
                type: 'google-terrain',
                module: 'gm',
                group: 'Google',
                template: 'gm/gmaps-tpl.html',
                name: IMatchTranslate.getStr('PROVIDER_GM_TER'),
                maxZoom: 21,
                params: {
                    fields: 'id,gps,name',
                    imagesize: 'T'
                }
            }            
        ]
    }


    /////////////////////////////////////////////////////////////////////////////
    // Get the active map config
    this.getActiveMapConfig = function()
    {
        return self.getMapConfig(self.appConfig.mapType);
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Get the map config for the given key
    this.getMapConfig = function(type)
    {
        for (var i = 0; i < mapTypes.length; ++i) {
            if (mapTypes[i].type == type) {
                return mapTypes[i];
            }
        }
        return null;
    }


    /////////////////////////////////////////////////////////////////////////////
    // Initialize the menu at the top. Append all available map types.
    function initMenu()
    {
        // The currently selected map
        var mc = self.getMapConfig(self.appConfig.mapType);

        // Append all available maps to the configure menu map select
        var list = $('#setting-map-selector');
        list.text('');

        var group = '';
        mapTypes.forEach(function(m) {
            list.append(`<option value="${m.type}"${m.type == self.appConfig.mapType ? 'selected' : ''}>${m.name}</option>`);
        }, this);
        
        // Apply options to menu

        setCheck('#check-show-file-track',self.appConfig.showFileTrack);
        setCheck('#setting-show-images',self.appConfig.showImages);
        setCheck('#setting-show-filedata',self.appConfig.showFileData);
        setCheck('#setting-show-locations',self.appConfig.showLocations);
        setCheck('#setting-show-locations-loc-menu',self.appConfig.showLocations);
        setCheck('#setting-follow-focus',self.appConfig.followFocus);

        $(`#form-settings-fov input[type=radio][name=fov][value=${self.appConfig.fieldOfView}]`).prop('checked',true);

        $('#map-refresh').click(function() {
            IMatch.refresh();
        });

        $('#map-refresh-settings').click(function() {
            IMatch.refresh();
        });

        $('#map-camdb-reset').click(function() {
            var m = new IMatchModals();
            m.showMessageBoxYesNo({
                headline: IMatchTranslate.getStr('MSGBOX_CAMDB_RESET_TITLE'),
                content: IMatchTranslate.getStr('MSGBOX_CAMDB_RESET_TEXT'),
                okCallback: function() {
                    self.camDB.reset();
                }
            });
        });
        

        // FOV form
        $('#form-settings-fov input[type=radio][name=fov]').change(function() {
            self.appConfig.fieldOfView = $(this).val();
            self.saveConfig();
            reloadMap();
           
            event.preventDefault();
            return false;
        });

        // The map selector list.
        $('#setting-map-selector').click(function(event) {
            // Avoid closing the menu when the user clicks into the control
            event.preventDefault();
            return false;
        });
        $('#setting-map-selector').on('change',function(event) {
            // Changed the map
            onMapChanged($(this).val());
            // Close the dropdown menu after choosing a map.
            $('#btn-settings').dropdown('toggle');
            event.preventDefault();
            return false;
        });

        // Other handlers for the Configure menu

        $('#setting-show-images').click(function(){
            self.appConfig.showImages = !self.appConfig.showImages;
            setCheck('#'+this.id,self.appConfig.showImages);
            self.saveConfig();
            reloadMap();
        });

        $('#setting-show-filedata').click(function(){
            self.appConfig.showFileData = !self.appConfig.showFileData;
            setCheck('#'+this.id,self.appConfig.showFileData);
            self.saveConfig();
            reloadMap();
        });

        $('#setting-show-locations').click(function(){
            self.appConfig.showLocations = !self.appConfig.showLocations;
            setCheck('#'+this.id,self.appConfig.showLocations);
            self.saveConfig();
            reloadMap();
        });
        $('#setting-show-locations-loc-menu').click(function(){
            self.appConfig.showLocations = !self.appConfig.showLocations;
            setCheck('#'+this.id,self.appConfig.showLocations);
            self.saveConfig();
            reloadMap();
        });

        $('#setting-follow-focus').click(function(){
            self.appConfig.followFocus = !self.appConfig.followFocus;
            setCheck('#'+this.id,self.appConfig.followFocus);
            self.saveConfig();
        });

        // If the database is read-only, we need to disable some commands

        if (self.dbReadOnly) {
            $('#btn-apply-target').hide();
        }
        else {
            $('#btn-apply-target').click(function(e) {
                if (mapModule) {
                    mapModule.onApplyTargetMarker();
                }
            });
        }

        // Button handlers

        // Zoom to extent (show all files on the map)
        $('#btn-zoom-extent').click(function() {
            closeModals();
            if (mapModule) {
                mapModule.onZoomToExtent();
            }
        });

        // Locate the target marker
        $('#btn-locate-target').click(function() {
            closeModals();
            if (mapModule) {
                mapModule.onLocateTargetMarker();
            }
        });

        // Open the find overlay
        $('#btn-find').click(function() {
            closeModals();
            var f = new MapFind();
            f.open();
        });

        // Open the edit location dialog
        $('#btn-location-edit').click(function() {
            closeModals();
            IMatch.editGeoLocation({});
        });

        // Display all locations
        $('#btn-location-goto').click(function() {
            closeModals();
            var f = new MapLocations();
            f.open();
        });

        // Trigger KML Export
        $('#btn-kml-export').click(function() {
            closeModals();
            var f = new KMLExport();
            f.open();
        });


        // If the lock button is clicked we reload the map and toggle the icon.
        $('#btn-lock').click(function() {
            self.appConfig.lockMove = !self.appConfig.lockMove;
            setLockBtnIcon();
            if (mapModule) {
                mapModule.onLockChanged();
            }
            self.saveConfig();
        });


        // Undo handler
        $('#btn-undo').click(function() {
            self.undo();
        });

        // The user clicked into the alert box about missing cameras.
        $('#btn-pending-cameras').click(event => {
            updateCameraDatabase();
            event.preventDefault();
            return false;
        });
        

        // Initially set the lock button icon
        setLockBtnIcon();

        initSelectTool();

        initTrackTool();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Set up and manage the select tool
    function initSelectTool()
    {
        // Set the initial toolbar scope
        $('#select-tb-where').val(self.appConfig.selectScope);

        // If the user changes the scope in the select toolbar, 
        // we save this setting.
        $('#select-tb-where').change(function(e) {
            self.appConfig.selectScope = this.value;
            self.saveConfig();
        });

        // The various shapes for the select tool
        $('#select-circle').click(function() {
            self.appConfig.selectTool = 'circle';
            closeModals();
            mapModule.onSelectTool();
        });

        $('#select-rect').click(function() {
            self.appConfig.selectTool = 'rectangle';
            closeModals();
            mapModule.onSelectTool();
        });
        
        $('#select-poly').click(function() {
            self.appConfig.selectTool = 'polygon';
            closeModals();
            mapModule.onSelectTool();
        });

        // The select toolbar, shown while the select tool is active,

        $('#btn-tb-select-apply').click(function() {
            mapModule.onSelectToolApply(true);
        });
        $('#btn-tb-select-cancel').click(function() {
            mapModule.onSelectToolApply(false);
        });

        setCheck('#select-use-shown',self.appConfig.selectToolUseShownCoordinates);

        $('#select-use-shown').click(function(){
            self.appConfig.selectToolUseShownCoordinates = !self.appConfig.selectToolUseShownCoordinates;
            setCheck('#'+this.id,self.appConfig.selectToolUseShownCoordinates);
            self.saveConfig();

            // If a select is currently active, re-run it as needed.
            if (toolbar == 'select') {
                closeModals();
                mapModule.onSelectTool();
            }
        });

    }


    /////////////////////////////////////////////////////////////////////////////
    // Set up and manage the track tool
    function initTrackTool()
    {
        updateTrackTool();

        // Load a track file and display the track on the map
        $('#btn-track-load').click(function() {
            loadGPXFile();
        });

        // Load all GPX files in a user-selected folder
        $('#btn-track-load-folder').click(function() {
            loadGPXTrackLogsFromFolder();
        });


        $('#check-show-file-track').click(function() {
            self.appConfig.showFileTrack = !self.appConfig.showFileTrack;
            mapModule.enableShowFileRoute(self.appConfig.showFileTrack);
            setCheck('#check-show-file-track',self.appConfig.showFileTrack);
            self.saveConfig();
        });
        
        // Center the map and show the full track
        $('#btn-track-view').click(function() {
            if (tracks.hasTracks()) {
                mapModule.onCenterTrack();
            }
        });

        // Show the configuration dialog
        $('#btn-track-configure').click(function() {
            closeModals();
            var f = new TrackLogConfig();
            f.open();
        });

        // Unload the current track
        $('#btn-track-unload').click(function() {
            if (tracks.hasTracks()) {
                tracks.clear();
                mapModule.onTrackUnLoaded();
            }
            updateTrackTool();
        });

        // Map files in the scope to the track and allows the user to view these (temporary) markers.
        $('#btn-track-select').click(function() {
            closeModals();
            trackSelectFiles();
        });

        // Map and geo-code all files
        $('#btn-track-geocode').click(function() {
            closeModals();
            trackGeoCodeFiles();
        });
    }

    function getSelectedFiles()
    {
        return IMWS.get('v1/files',{
            idlist: IMatch.idlist.fileWindowSelection,
            fields: 'id,filename,folder',
        });
    }

    /////////////////////////////////////////////////////////////////////////////
    // Load a GPX file.
    async function loadGPXFile()
    {
        // If the option to remember the last folder is set, we check
        // for the last loaded GPX file and use it to initialize the dialog.
        var lastfile = '';
        var logfiles = self.appConfig.trackLog.files;
        if (self.appConfig.trackLog.rememberFolder) {
            if (logfiles.length > 0) {
                lastfile = logfiles[logfiles.length-1];
            }
        }
        else {
            // Else we use the first selected file as the initial folder.
            var selection = await getSelectedFiles();
            if (selection && selection.files.length) {
                lastfile = selection.files[0].fileName;
            }
        }

        // Show the file browser
        IMatch.fileBrowserFS({
            type: 'open',
            title: IMatchTranslate.getStr('TRACK_DLG_LOAD_PROMPT'),
            filter: IMatchTranslate.getStr('TRACK_DLG_FILEMASK'),
            initialdir: lastfile,
            filename: lastfile,
        }).then(function(response) {
            if (response.result == 'ok') {

                // Update logfiles with the new file name.
                // We first check if the file name is already in logfiles 
                // and remove it.
                if (self.appConfig.trackLog.rememberFolder) {
                    for (var i = 0; i < logfiles.length; ++i) {
                        if (logfiles[i] == response.fileName) {
                            logfiles.splice(i,1);
                            break;
                        }
                    }
                    // Limit the number of last files to 5
                    if (logfiles.length >= 5) {
                        logfiles.splice(0,1);
                    }
                    logfiles.push(response.fileName);
                }

                tracks.loadTrackLog(response.fileName).then(function() {
                    updateTrackTool();
                    if (!tracks.hasTracks()) {
                        var m = new IMatchModals();
                        m.showMessageBoxOK({
                            headline: IMatchTranslate.getStr('MSGBOX_TRACK_TOOL'),
                            content: IMatchTranslate.getStr('MSGBOX_TRACK_IMPORT_EMPTY')
                        });
                    }
                    else {
                        trackSelectFiles();
                    }
                    // Even when there are no points!
                    mapModule.onTrackLoaded(true);
                },function(error) {
                    var m = new IMatchModals();
                    m.showMessageBoxOK({
                        headline: IMatchTranslate.getStr('MSGBOX_TRACK_TOOL'),
                        content: IMatchTranslate.getStr('MSGBOX_TRACK_IMPORT_ERROR_PARSE').replace('{0}',error)
                    });
                });
            }
        });
    }

    /////////////////////////////////////////////////////////////////////////////
    // Load all GPX files in the selected folder.
    async function loadGPXTrackLogsFromFolder()
    {

        // If the option to remember the last folder is set, we check
        // for the last loaded GPX file and use it to initialize the dialog.
        var lastfolder = '';
        if (self.appConfig.trackLog.rememberFolder) {
            var folders = self.appConfig.trackLog.folders;
            if (folders.length > 0) {
                lastfolder = folders[folders.length-1];
            }
        }
        else {
            // Else we use the first selected file as the initial folder.
            var selection = await getSelectedFiles();
            if (selection && selection.files.length) {
                lastfolder = selection.files[0].folder;
            }
        }

        // Show the file browser
        IMatch.folderBrowserFS({
            title: IMatchTranslate.getStr('TRACK_FOLDER_DLG_LOAD_PROMPT'),
            initialdir: lastfolder
        }).then(function(response) {
            if (response.result == 'ok') {

                // Remember this folder
                if (self.appConfig.trackLog.rememberFolder) {
                    for (var i = 0; i < folders.length; ++i) {
                        if (folders[i] == response.folderName) {
                            folders.splice(i,1);
                            break;
                        }
                    }
                    // Limit the number of last files to 5
                    if (folders.length >= 5) {
                        folders.splice(0,1);
                    }
                    folders.push(response.folderName);
                }

                tracks.loadTrackLogsFromFolder(response.folderName).then(() => {
                    updateTrackTool();
                    if (!tracks.hasTracks()) {
                        var m = new IMatchModals();
                        m.showMessageBoxOK({
                            headline: IMatchTranslate.getStr('MSGBOX_TRACK_TOOL'),
                            content: IMatchTranslate.getStr('MSGBOX_TRACK_IMPORT_EMPTY')
                        });
                    }
                    else {
                        trackSelectFiles();
                    }
                    // Even when there are no points!
                    mapModule.onTrackLoaded(true);
                    return;
                });
            }
        });        
    }


    /////////////////////////////////////////////////////////////////////////////
    // Enable/disable commands
    function updateTrackTool()
    {
        if (tracks.hasTracks()) {
            $('#btn-track-select').parent().removeClass('disabled');
            $('#btn-track-geocode').parent().removeClass('disabled');
            $('#btn-track-unload').parent().removeClass('disabled');
            $('#btn-track-view').parent().removeClass('disabled');
        }
        else {
            $('#btn-track-select').parent().addClass('disabled');
            $('#btn-track-geocode').parent().addClass('disabled');
            $('#btn-track-unload').parent().addClass('disabled');
            $('#btn-track-view').parent().addClass('disabled');
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Set the lock button icon
    function setLockBtnIcon() {
        var btn = $('#btn-lock-icon');
        if (self.appConfig.lockMove) {
            btn.removeClass('fa-unlock');
            btn.addClass('fa-lock');
        }
        else {
            btn.removeClass('fa-lock');
            btn.addClass('fa-unlock');
        }
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // switch the toolbar
    this.switchToolbar = function(toolbar)
    {
        if (toolbar != currentToolbar) {
            currentToolbar = toolbar;
            if (toolbar == 'default') {
                $('#toolbar-select').addClass('tb-hidden');
                $('#toolbar-default').removeClass('tb-hidden');
            }
            else if (toolbar == 'select') {
                $('#toolbar-default').addClass('tb-hidden');
                $('#toolbar-select').removeClass('tb-hidden');
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Show or hide the undo button
    this.setUndo = function(enable) 
    {
        if (enable) {
            $('#btn-undo').fadeIn();
        }
        else {
            $('#btn-undo').fadeOut();
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // This is called by the map module when the target marker is added or removed.
    this.onTargetMarkerChanged = function(show)
    {
        $('#btn-locate-target').attr('disabled',!show);
    }


    /////////////////////////////////////////////////////////////////////////////
    // Bring the given location into view. 
    this.gotoLocation = function(guid) {
        if (mapModule) {
            mapModule.onGotoLocation(guid);
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Called by overlays
    self.setActiveOverlay = function (name) 
    {
        activeOverlay = name;
    }


    /////////////////////////////////////////////////////////////////////////////
    // Returns either the Google Maps key provided by the user (preferred!)
    // or the shaky shared key.
    self.getGMapsKey = function()
    {
        if (self.IMatchConfig.gmapsKey == '') {
            return '';
        }
        else {
            return self.IMatchConfig.gmapsKey;
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // When the browser winsow is resized, we reduce the size of the buttons.
    $(window).resize(function() {
        var ww = window.innerWidth;
        if (ww < 300) {
            $('#menu-bar .btn').each(function(b) {
                $(this).addClass('btn-sm');
            });
        }
        else {
            $('#menu-bar .btn').each(function(b) {
                $(this).removeClass('btn-sm');
            });
        }
    });


    /////////////////////////////////////////////////////////////////////////////
    // Set or un-set a check box
    // This assumes that the element with id has a child element with the ptc-checkbox class.
    function setCheck(id,on)
    {
        var t = $(id + ' .ptc-checkbox');
        t.removeClass('fa fa-check-square-o fa-lg');
        t.removeClass('fa fa-square-o fa-lg text-muted');
        t.addClass(on ? 'fa fa-check-square-o fa-lg' : 'fa fa-square-o fa-lg text-muted');
    }


    /////////////////////////////////////////////////////////////////////////////
    // This is called when the user chooses a different map in the configure menu.
    function onMapChanged(type)
    {
        closeModals();

        var mc = self.getMapConfig(type);

        // Depending on the group, we may need a key
        if (mc.group == 'Google') {
            // We need a key for this. If no key is configured, inform the user.
            if (self.IMatchConfig.gmapsKey == '') {
                // Handled in loadMapTemplate!
                loadMapTemplate(mc);
            }
            else {
                loadMapTemplate(mc);
            }
        }
        else if (mc.group == 'Bing') {
            googleSharedKeyWarningShown = false;
            // We need a key for this. If no key is configured, inform the user.
            if (self.IMatchConfig.bingMapsKey == '') {
                var m = new IMatchModals();
                m.showMessageBoxOK({
                    headline: IMatchTranslate.getStr('MSGBOX_NO_BING_KEY_TILE'),
                    content: IMatchTranslate.getStr('MSGBOX_NO_BING_KEY_TEXT')
                });
                return;
            }
            else {
                loadMapTemplate(mc);
            }
        }
        else if (mc.group == 'HERE') {
            googleSharedKeyWarningShown = false;
            // We need a key for this. If no key is configured, inform the user.
            if (self.IMatchConfig.HEREMapsKey == '') {
                var m = new IMatchModals();
                m.showMessageBoxOK({
                    headline: IMatchTranslate.getStr('MSGBOX_NO_HERE_KEY_TILE'),
                    content: IMatchTranslate.getStr('MSGBOX_NO_HERE_KEY_TEXT')
                });
                return;
            }
            else {
                // IMatch 2020
                // We support old API codes with a:b and also API keyss (one string)
                loadMapTemplate(mc);
                // let key = self.IMatchConfig.HEREMapsKey.split(':');
                // if (key.length < 2) {
                //     var m = new IMatchModals();
                //     m.showMessageBoxOK({
                //         headline: IMatchTranslate.getStr('MSGBOX_INVALID_HERE_KEY_TILE'),
                //         content: IMatchTranslate.getStr('MSGBOX_INVALID_HERE_KEY_TEST')
                //     });
                //     return;
                // }
                // else {
                //     loadMapTemplate(mc);
                // }
            }
        }
        else {
            googleSharedKeyWarningShown = false;
            loadMapTemplate(mc);
        }

        self.appConfig.mapType = type;

        // Update the menu
        mapTypes.forEach(function(m) {
            setCheck('#'+m.type,m.type == self.appConfig.mapType);
        });

        self.saveConfig();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Reloads the map after settings have changed
    function reloadMap()
    {
        closeModals();

        var mc = self.getActiveMapConfig();
        loadMapTemplate(mc);
    }

    /////////////////////////////////////////////////////////////////////////////
    // Loads the specified map template and replaces the contents of the #map-container
    function loadMapTemplate(mapData)
    {
        if (mapModule) {
            mapModule.onDestroy();
        }
        mapModule = null;

        $( "#map-container" ).hide().load(mapData.template, function() {
            if (mapData.module == 'ol') {

                // We don't support rectangle selection for OL. OL has no rectangles.
                $('#select-rect').hide();

                mapModule = new OLMapModule();
                mapModule.enableShowFileRoute(self.appConfig.showFileTrack);
                mapModule.loadMap(true);
            }
            else if (mapData.module == 'gm') {

                // We support rectangles for GM
                $('#select-rect').show();

                if (self.IMatchConfig.gmapsKey == '') {
                    if (!googleSharedKeyWarningShown) {
                        googleSharedKeyWarningShown = true;
                        var m = new IMatchModals();
                        m.showMessageBoxOK({
                            headline: IMatchTranslate.getStr('MSGBOX_TEST_GMAPS_KEY_TILE'),
                            content: IMatchTranslate.getStr('MSGBOX_TEST_GMAPS_KEY_TEXT')
                        });
                    }
                }

                mapModule = new GMapModule();
                mapModule.enableShowFileRoute(self.appConfig.showFileTrack);
                mapModule.loadMap(true);
            }
        }).fadeIn(500);
    }

    
    // Close all open modal windows and overlays.
    // Also closes popups shown by maps.
    function closeModals()
    {
        var m = new IMatchModals();
        m.closeModal();
        
        var mm = new MapModals();
        mm.closeAll();

        if (mapModule !== null) {
            mapModule.onClosePopups();
        }
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Create the URL for the image with the given id and size. 
    // If size is not specified, we use 'T'.
    this.fileUrl = function (id,size) {
        var size = size || 'T';
        // We always load the default image if the server cannot create an image for the requested file
        return IMatch.IMWSUrl() + '/v1/files/image?auth_token=' + IMatch.authToken() + '&id=' + id + '&imagesize=' + size + '&default=true';
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Returns the URL of the 'shield' image for protected files.
    this.shieldIconUrl = function (vector) {
        vector = vector || false;
        if (vector) {
            return 'img/shield.svg';
        }
        else {
            return 'img/shield.png';
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Displays an error message to the user.
    // This is called by various sub-modules    
    this.showError = function(type,error)
    {   
        window.setTimeout(() => {
            let params = {
                icon: 'fa fa-frown fa-3x',
            }

            if (type == 'setGps') {
                params.headline = IMatchTranslate.getStr('ERROR_SETGPS_TITLE');
                params.content = IMatchTranslate.getStr('ERROR_SETGPS_TEXT');
                if (error.responseJSON != undefined) {
                    params.additionalInfo = '<pre>' + JSON.stringify(error.responseJSON,null,2) + '</pre>';
                }
            }
            else {
                params.headline = IMatchTranslate.getStr('ERROR_UNKNOWN_TITLE');
                params.content = IMatchTranslate.getStr('ERROR_UNKNOWN_TEXT');
                if (error.responseJSON != undefined) {
                    params.additionalInfo = JSON.stringify(error.responseJSON,null,2);
                    params.additionalInfo = '<pre>' + JSON.stringify(error.responseJSON,null,2) + '</pre>';
                }
            }

            closeModals();
            let m = new IMatchModals();
            m.showErrorBox(params);
            console.error(params);
        },1000);
    }

    /////////////////////////////////////////////////////////////////////////////
    // Opens a result window for the given files. 
    // If there is already a result window for FindNearby open, it is reused.
    this.openFindNearbyResultWindow = function(name,id)
    {
        if (findResultWindowInstanceId == 0) {

            // if id is empty, we display a message to the user
            if (id.length == 0) {
                var m = new IMatchModals;
                m.showMessageBoxOK({
                    headline: IMatchTranslate.getStr('MSGBOX_FINDNEARBY_NORESULT_TITLE'),
                    content: IMatchTranslate.getStr('MSGBOX_FINDNEARBY_NORESULT_TEXT')
                });
                return;
            }

            IMatch.openResultWindow({
                name: name,
                id: JSON.stringify(id)
            }).then(function(response) {
                findResultWindowInstanceId = response.instanceId;
            });
        }
        else {
            // Reuse the existing result window if it still exists (may have been closed by the user).
            // We also update it when nothing was found!
            IMatch.updateResultWindow({
                instanceid: findResultWindowInstanceId,
                id: JSON.stringify(id)
            }).then(function(response) {
                // Success
            },error => {
                // Does not longer exist
                findResultWindowInstanceId = 0;
                self.openFindNearbyResultWindow(name,id);
                return;
            });

            // if id is empty, we display a message to the user
            if (id.length == 0) {
                var m = new IMatchModals;
                m.showMessageBoxOK({
                    headline: IMatchTranslate.getStr('MSGBOX_FINDNEARBY_NORESULT_TITLE'),
                    content: IMatchTranslate.getStr('MSGBOX_FINDNEARBY_NORESULT_TEXT')
                });
            }
        }
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Select all files in the current scope which are on 
    // the currently loaded track.
    function trackSelectFiles()
    {
        if (tracks.hasTracks()) {
            tracks.mapFiles().then(function() {
                var m = new IMatchModals;
                var msg;
                if (tracks.trackData().filesMapped == 0) {
                    // No result
                    msg = IMatchTranslate.getStr('MSGBOX_TRACK_TOOL_MAP_RESULT_NONE');
                }
                else {
                    msg = IMatchTranslate.getStr('MSGBOX_TRACK_TOOL_MAP_RESULT');
                }
                msg = msg.replace('{0}',tracks.trackData().filesMapped);
                msg = msg.replace('{1}',tracks.trackData().filesToMap);
                msg = msg.replace('{2}',tracks.trackData().filesIgnored);

                if (tracks.trackData().filesMapped == 0) {
                    m.showMessageBoxOK({
                        headline: IMatchTranslate.getStr('MSGBOX_TRACK_TOOL'),
                        content: msg
                    });
                }
                else {
                    m.showMessageBoxYesNo({
                        headline: IMatchTranslate.getStr('MSGBOX_TRACK_TOOL'),
                        content: msg,
                        okCallback: function() {
                            mapModule.onShowTrackFiles();
                        }
                    });
                }
            });
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Geo-code files
    // We setup the geoCodeTaskData.tasks with the data for all files.
    // Then we show the progress overlay and start the processing. We process the
    // files in batches for best speed.
    function trackGeoCodeFiles()
    {
        if (tracks.hasTracks()) {
            tracks.mapFiles().then(function() {
                geoCodeTaskData.running = true;
                geoCodeTaskData.index = 0;
                geoCodeTaskData.tasks = [];
                geoCodeTaskData.filesUpdated = [];

                // For each mapped file, add the update tasks 
                let filesToUpdate = 0;
                for (var i = 0; i < tracks.trackData().files.length; ++i) {
                    var f = tracks.trackData().files[i];
                    if (f.mapped) {
                        filesToUpdate++;
                        geoCodeTaskData.tasks.push({
                            id: f.id,
                            tag: 'gpslatitude',
                            op: 'set',
                            rawvalue : true,
                            value: f.mappedGPS.lat
                        });

                        geoCodeTaskData.tasks.push({
                            id: f.id,
                            tag: 'gpslongitude',
                            op: 'set',
                            rawvalue : true,
                            value: f.mappedGPS.lng
                        });

                        // geoCodeTaskData.tasks.push({
                        //     id: f.id,
                        //     tag: 'gpsdestlatitude',
                        //     op: 'set',
                        //     rawvalue : true,
                        //     value: f.mappedGPS.lat
                        // });

                        // geoCodeTaskData.tasks.push({
                        //     id: f.id,
                        //     tag: 'gpsdestlongitude',
                        //     op: 'set',
                        //     rawvalue : true,
                        //     value: f.mappedGPS.lng
                        // });

                        geoCodeTaskData.tasks.push({
                            id: f.id,
                            tag: 'gpsaltitude',
                            op: 'set',
                            rawvalue : true,
                            value: f.mappedGPS.alt
                        });

                        geoCodeTaskData.tasks.push({
                            id: f.id,
                            tag: 'gpsdatetime',
                            op: 'set',
                            rawvalue : false,
                            value: moment.utc(f.mappedGPS.time).format()
                        });
                    }
                }

                if (geoCodeTaskData.tasks.length > 0) {
                    // console.log('Updating GPS coordinates for ' + filesToUpdate + ' files.');
                    geoCodeBusy(true);
                }
            });
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Show the progress bar overlay
    function geoCodeBusy(busy)
    {
        if (busy) {
            if (geoCodeTaskData.overlay == null) {
                geoCodeTaskData.overlay = new MapModals;
            }
            geoCodeTaskData.overlay.showProgressOverlay(true,
            function() {
                // Overlay created. Start processing.
                geoCodeTaskData.overlay.setProgressOverlayPercentage(0,IMatchTranslate.getStr('MSGBOX_TRACK_GEOCODE_PROGRESS_TITLE'));
                geoCodeNextBatch();
            },
            function() {
                // Aborted
                geoCodeTaskData.running = false;
                geoCodeBusy(false);
            });
        }
        else {
            geoCodeTaskData.running = false;
            if (geoCodeTaskData.overlay) {
                geoCodeTaskData.overlay.closeProgressOverlay();
            }
            
            mapModule.onHideTrackFiles();
            
            // Select all updated files. This makes the markers visible on the map.
            if (geoCodeTaskData.filesUpdated.length > 0) {
                IMatch.selectFiles({
                    id: geoCodeTaskData.filesUpdated.join(',')
                });
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function geoCodeNextBatch()
    {
        if (!geoCodeTaskData.running) return;

        if (geoCodeTaskData.index < geoCodeTaskData.tasks.length) {

            geoCodeTaskData.overlay.setProgressOverlayPercentage((geoCodeTaskData.index*100)/Math.max(1,geoCodeTaskData.tasks.length));

            // We process the tasks in batches of a multiple of 4 (4 tasks per file):
            var tasks = [];
            while (geoCodeTaskData.index < geoCodeTaskData.tasks.length) {
                tasks.push(geoCodeTaskData.tasks[geoCodeTaskData.index]);
                // Remember the files we have updated.
                geoCodeTaskData.filesUpdated.push(geoCodeTaskData.tasks[geoCodeTaskData.index].id);
                geoCodeTaskData.index++;
                if (tasks.length >= (10 * 4)) break;
            }

            var params = {
                tasks : JSON.stringify(tasks)
            }

            // Let IMatch do the work for us...
            IMWS.post('v1/metadata',params).then(function(response) {
                geoCodeNextBatch();
            },
            function (error) {
                geoCodeTaskData.running = false;
                geoCodeBusy(false);
            });
        }
        else {
            geoCodeTaskData.running = false;
            geoCodeBusy(false);
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Display a modal with info about missing cameras and let the user send
    // an email
    function updateCameraDatabase()
    {
        // Get the pending cameras from the camera database and compose an email.
        let msg = IMatchTranslate.getStr('CAMDB_MISSING_INFO');
        for (let c of self.camDB.getPendingCameras()) {
            let l = IMatchTranslate.getStr('CAMDB_MISSING_TEMPLATE');
            l = l.replace('{make}',c.make);
            l = l.replace('{model}',c.model);
            msg += l;
        }

        IMatch.clipboardCopy({
            format : 'text',
            text : msg
        });
        
        let m = new IMatchModals();
        m.showMessageBoxYesNo({
            headline: IMatchTranslate.getStr('MSGBOX_CAMDB_MISSING_TITLE'),
            content: IMatchTranslate.getStr('MSGBOX_CAMDB_MISSING_TEXT'),
            okCaption: IMatchTranslate.getStr('MSGBOX_CAMDB_BTN_SEND'),
            // Make it as wide as possible
            modalClass: 'full-width',
            okCallback: function() {

                // The embedded browser (Chromium) in IMatch does not support mailto: 
                // Hence we use ShellExecute in the special mailto: syntax to create the email:

                let params = `"mailto:support@photools.com?subject=Camera Database Update&body=${encodeURI(msg)}"`;
                IMatch.shellExecute({
                    executable : params,
                    verb : null,
                    parameters : null
                });
            }
        });
   
    }


    /////////////////////////////////////////////////////////////////////////////
    // This is called by the map modules after loading new files.
    // Checks the camera database has reported files with missing data.
    // We then inform the user, once per session.
    self.checkCameraDB = function()
    {
        if (!cameraDBWarningShown && (self.appConfig.fieldOfView != 'off')) {
            window.setTimeout(function() {
                if (self.camDB.hasPending()) {
                    cameraDBWarningShown = true;
                    $('#camdb-warning-container').fadeIn();
                }
            },1000);
        }
    }
    
}  // app
